xquery version "3.1";

module namespace elastic="http://elastic.io";

import module namespace config="https://sade.textgrid.de/ns/config" at "config.xqm";
import module namespace console="http://exist-db.org/xquery/console";
(:~ import module namespace xdmp="http://marklogic.com/xdmp";  ~:)
import module namespace functx = 'http://www.functx.com';
import module namespace lesapi="http://sepia.io/esapi/legacy-api";
import module namespace transformKMG="http://mollenhauer-edition.de/ns/transform" at "./kmg/transform.xqm";


declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xhtml="http://www.w3.org/1999/xhtml";
declare namespace http="http://expath.org/ns/http-client";

declare namespace elastic="http://elastic.io";

declare variable $elastic:uri := 'http://127.0.0.1:9200';
declare variable $elastic:prefix := 'at_';

(:
  Defines generic functions for searching, indexing and rebuilding several
  elasticsearch indices. The elasticsearch uri can be configured above, you may
  also set a prefix for index names in case the namespace is shared.

  elastic:setup() serves to (re)build all elasticsearch indices from register
  files, edition pages and wiki pages. Changes to wiki pages are immediately
  re-indexed, see /modules/wiki/confluence.xqm.
:)

declare function elastic:index-person($person) {
    let $doc := map {
        "id": substring-after($person/@ref, ":"),
        "name": map {
            "de": elastic:textFrom($person/tei:persName[@xml:lang='de']),
            "fr": elastic:textFrom($person/tei:persName[@xml:lang='fra'])
        },
        "search_data": map {
            "de": elastic:textFrom($person/*[@xml:lang='de'] | $person/*[not(@xml:lang)]),
            "fr": elastic:textFrom($person/*[@xml:lang='fra'] | $person/*[not(@xml:lang)])
        }
    }
    return lesapi:index-doc($elastic:uri, $elastic:prefix || 'people', $doc)
};


declare function elastic:index-work($work) {
    let $doc := map {
        "id": substring-after($work/@ref, ":") => replace("\..*", ""),
        "name": map {
            "de": elastic:textFrom($work/tei:name[@xml:lang='de']),
            "fr": elastic:textFrom($work/tei:name[@xml:lang='fra'])
        },
        "search_data": map {
            "de": elastic:textFrom($work/*[@xml:lang='de']),
            "fr": elastic:textFrom($work/*[@xml:lang='fra'])
        }
    }
    return lesapi:index-doc($elastic:uri, $elastic:prefix || 'works', $doc)
};

declare function elastic:index-edition($doc, $locale) {
    let $witnesses :=  $doc//tei:sourceDesc/tei:listWit/tei:witness

    let $tg_uri := fn:tokenize(util:document-name($doc),"[-\.]")[1]
    (:~ let $page_number := 2 ~:)

    let $editionNode := $doc/tei:TEI/tei:text[not(@type='commentary')]

    let $name := string($doc//tei:title[@type="main"]/text())
    let $id := concat('edition-', $tg_uri)
    let $doc_map := map {
        "id": $id,
        "edition": $tg_uri,
        (:~ "page_number": $page_number, ~:)
        "locale": $locale,
        "name": $name
    }
    let $version_maps := elastic:versionMap($doc, $editionNode, $witnesses)
    let $version_maps_merged := map:merge($version_maps)
    let $doc_map := map:merge(($doc_map, $version_maps_merged))

    return
        lesapi:index-doc($elastic:uri, $elastic:prefix || 'editions', $doc_map)
};

declare function elastic:index-commentary($doc, $locale) {
    let $tg_uri := fn:tokenize(util:document-name($doc),"[-\.]")[1]
    let $edition_commentary := elastic:textFrom($doc/tei:TEI/tei:text[@type='commentary'])
    let $name := string($doc//tei:title[@type="main"]/text())
    let $id := concat('edition-', $tg_uri)
    let $doc := map {
        "id": $id,
        "edition": $tg_uri,
        "locale": $locale,
        "name": $name,
        "search_data": fn:normalize-space($edition_commentary)
    }
    return
        lesapi:index-doc($elastic:uri, $elastic:prefix || 'commentaries', $doc)
};


declare function elastic:index-wiki-page($page, $locale) {
    let $uri := string(base-uri($page))
    let $doc_name := fn:tokenize($uri, '/')[6]
    let $parts := fn:tokenize($doc_name, '[_\.]')
    let $id := $parts[1]
    let $locale := $parts[2]
    let $elastic_doc_id := concat($id, '_', $locale)
    let $doc := map {
        "uri": $uri,
        "id": $id,
        "locale": $locale,
        "name": string($page//h1[1]/span[1]),
        "search_data": fn:normalize-space(elastic:textFrom($page))
    }
    return lesapi:index-doc($elastic:uri, $elastic:prefix || 'wiki_pages', $elastic_doc_id, $doc)
};

declare function elastic:require_ok($response) {
    let $code := xs:integer($response[1]/@status)
    return 
        if ($code < 200 or $code > 299) then
            let $body := util:base64-decode($response[2])
            let $name := QName('http://datypic.com/err', 'ProdNumReq')
            let $msg := concat("(", string($code), ") elasticsearch responded with:", $body)
            return fn:error($name, $msg)
        else
            fn:true()
};




declare function elastic:versionMap($doc, $editionNode, $witnesses) {
    let $log := util:log-system-out("Indexing document: " || util:document-name($doc))
    for $witness in $witnesses
    let $version := string($witness/@xml:id)
    let $StringNormalized := elastic:getVersionString($doc, $editionNode, $version)
    let $version_map := map{"search_data_" || $version: $StringNormalized}
    return $version_map
};


declare function elastic:getVersionString($doc, $editionNode, $version) {

    let $htmlNode := transformKMG:transform($editionNode, util:document-name($doc), $version)
    let $log := util:log-system-out("Indexing version " || $version)
    let $htmlNode := elastic:transform($htmlNode)

    let $StringNormalized := elastic:concat-normalize-string($htmlNode)
    let $log := util:log-system-out("Indexing finished for version " || $version)
    return $StringNormalized
};


declare function elastic:transform($editionNode){
    for $node in $editionNode
    return
        try {
             elastic:filterHTMLNodes($node)
        }
        catch * {
            util:log-system-out(util:system-time() || " error")
        }
};

declare function elastic:filterHTMLNodes($node)
{
  typeswitch($node)
    case element(xhtml:div) return
        if ( string($node/@class) =  "modal-body__content" )
        then 
        ()
        else    
            elastic:transform($node/node())
    case text() return
        text { $node }
        
    default return 
    elastic:transform($node/node())
};


declare function elastic:textFrom($e) {
    let $raw := string-join($e/descendant-or-self::*/text(), ' ')
    let $fixed_punctuation := replace($raw, ' ([\\.\\,])', '$1')
    return $fixed_punctuation
};

declare function elastic:concat-normalize-string($strings) {
    let $raw := string-join($strings, '') 
    let $edition_text := fn:normalize-space($raw)
    return $edition_text
};


declare function elastic:to_json($data) {
  let $result := fn:serialize($data, map { 'method': 'json'})
  return $result
};

declare function elastic:setup() {

    let $ editions_de := collection($config:sade-data)

    let $settings := map {
        "analysis": map {
            "filter": map {
                (: "german_keywords": map {
                "type":       "keyword_marker",
                "keywords":   ["Beispiel"]
                }, :)
                "german_stop": map {
                    "type":       "stop",
                    "stopwords":  "_german_"
                },
                "german_stemmer": map {
                    "type":       "stemmer",
                    "language":   "light_german"
                },
                "corrections": map {
                    "type": "synonym",
                    (:~ "synonyms": [] ~:)
                    "synonyms": ["Bewußtsein, Bewußstein"]
                    (: "synonyms": ["Bewußtstein => Bewußtstein, Bewußtsein"] :)
                    (: "synonyms_path": "synonyms.txt" :)
                }
            },
            "analyzer": map {
                "rebuilt_german": map {
                    "tokenizer": "standard",
                    "filter": [
                        "lowercase",
                        (:~ "german_stop", ~:)
                        (: "german_keywords", :)
                        "german_normalization",
                        "german_stemmer",
                        "corrections"
                    ]
                }
            }
        }
    }
    let $register_mappings := map {
        "_doc": map {
            "properties": map {
                "id": map {"type": "keyword", "index": fn:false()},
                "name": map {
                    "properties": map {
                        "de": map {"type": "text", "analyzer": "rebuilt_german"}
                    }
                },
                "search_data": map {
                    "properties": map {
                        "de": map {"type": "text", "analyzer": "rebuilt_german"}
                    }
                }
            }
        }
    }
    let $r := [
        lesapi:drop-index($elastic:uri, $elastic:prefix || "people"),
        lesapi:drop-index($elastic:uri, $elastic:prefix || "works"),
        lesapi:drop-index($elastic:uri, $elastic:prefix || "wiki_pages"),
        lesapi:drop-index($elastic:uri, $elastic:prefix || "edition_pages"),
        lesapi:drop-index($elastic:uri, $elastic:prefix || "editions"),
        lesapi:drop-index($elastic:uri, $elastic:prefix || "commentaries"),
        
        lesapi:create-index($elastic:uri, $elastic:prefix || "people", map {"settings": $settings, "mappings": $register_mappings}),
        lesapi:create-index($elastic:uri, $elastic:prefix || "works", map {"settings": $settings, "mappings": $register_mappings}),
        lesapi:create-index($elastic:uri, $elastic:prefix || "wiki_pages", map {"settings": $settings, "mappings": map {
            "_doc": map {
                "properties": map {
                    "id": map {"type": "keyword", "index": fn:false()},
                    "locale": map {"type": "keyword"},
                    "name": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data": map {"type": "text", "analyzer": "rebuilt_german"}
                }
            }
        }}),
        lesapi:create-index($elastic:uri, $elastic:prefix || "editions", map {"settings": $settings, "mappings": map {
            "_doc": map {
                "properties": map {
                    "id": map {"type": "keyword", "index": fn:false()},
                    "edition": map {"type": "keyword", "index": fn:false()},
                    (: "page_number": map {"type": "integer"}, :)
                    "locale": map {"type": "keyword"},
                    "name": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data": map {"type": "text", "analyzer": "rebuilt_german"},
                    (: "search_data": map {"type": "text", "analyzer": "rebuilt_german", "fields":}, :)
                    "search_data_a": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data_b": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data_c": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data_d": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data_e": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data_f": map {"type": "text", "analyzer": "rebuilt_german"}


                }
            }
        }}),
        lesapi:create-index($elastic:uri, $elastic:prefix || "commentaries", map {"settings": $settings, "mappings": map {
            "_doc": map {
                "properties": map {
                    "id": map {"type": "keyword", "index": fn:false()},
                    "edition": map {"type": "keyword", "index": fn:false()},
                    "page_number": map {"type": "integer"},
                    "locale": map {"type": "keyword"},
                    "name": map {"type": "text", "analyzer": "rebuilt_german"},
                    "search_data": map {"type": "text", "analyzer": "rebuilt_german"}
                }
            }
        }}),
        
        for $doc in collection($config:register)//tei:person
            return elastic:index-person($doc),
            
        for $doc in collection($config:register)//tei:item
            return elastic:index-work($doc),
    
        for $doc at $pos in $editions_de
            return elastic:index-edition($doc, "de"),

        for $doc at $pos in $editions_de
            return elastic:index-commentary($doc, "de"),
        
        for $doc in collection(concat($config:app-root, '/docs'))//div[@id='wiki']
            let $uri := string(base-uri($doc))
            let $doc_name := fn:tokenize($uri, '/')[6]
            let $parts := fn:tokenize($doc_name, '[_\.]')
            let $locale := $parts[2]
            let $revision := xs:integer(replace($parts[3], 'rev', ''))
            order by $revision
            return elastic:index-wiki-page($doc, $locale),
        
        lesapi:refresh($elastic:uri)
    ]
    
    return $r
};
