xquery version "3.1";

module namespace tei2plain="https://mollenhauer-edition.de/ns/conversions/tei2plain";

(:~  
 : Convert Module TEI to plain text
 : *******************************************************************************************
 : This module converts TEI-XML to plain text.
 :  
 : @version 1.0 (2021-09-23)
 : @status dev
 : @author Uwe Sikora, Max Zeterberg
 :)

import module namespace simplify="https://mollenhauer-edition.de/ns/conversions/simplify" at "simplify.xqm";
import module namespace removal="https://mollenhauer-edition.de/ns/conversions/removal" at "removal.xqm";
import module namespace qm="https://mollenhauer-edition.de/ns/conversions/quotation-marks" at "quotation-marks.xqm";

declare namespace tei = "http://www.tei-c.org/ns/1.0";
declare namespace output="http://www.w3.org/2010/xslt-xquery-serialization";

declare option output:method "text";

(:~
 : Converts the given node to plain text and inserts a carriage return after it.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $node: The current node
:)
declare function tei2plain:make-block($node as node()*) {
    tei2plain:newline($node/node()), '&#10;'
};

(:~
 : This functions puts carriage returns before block-elements and replaces tei:milestone/@type="paragraph" with carriage return.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $node: The current node
:)
declare function tei2plain:newline($nodes as node()*) {
    for $node in $nodes
    return
        if (local-name($node) = ('epigraph', 'formula', 'head', 'lg', 'list', 'p', 'seg', 'table')) then
            ('&#10;', tei2plain:convert($node))
        else
            if ((local-name($node) = 'milestone') and ($node/@type = 'paragraph')) then
                '&#10;'
            else
                tei2plain:convert($node)
};


(:~
 : This function converts the TEI-Document into plain text.
 :
 : @author Max-Ferdinand Zeterberg
 : @author Uwe Sikora
 : @version 1.0
 :  
 : @param $nodes: A pre-processed TEI-Document.
:)
declare function tei2plain:convert
($nodes as node()*) as item()* {
    for $node in $nodes
    return
        typeswitch ($node)
            case text()
                return
                    (normalize-space($node))
            case element(tei:closer)
                return
                    tei2plain:make-block($node)
            case element(tei:epigraph)
                return
                    tei2plain:make-block($node)
            case element(tei:formula)
                return
                    ('//TEX/', $node/node(), '/TEX//')
            case element(tei:head)
                return
                    tei2plain:make-block($node)
            case element(tei:lg)
                return
                    tei2plain:make-block($node)
            case element(tei:list)
                return
                    tei2plain:make-block($node)
            case element(tei:opener)
                return
                    tei2plain:make-block($node)
            case element(tei:p)
                return
                    tei2plain:make-block($node)
            case element(tei:seg)
                return
                    tei2plain:make-block($node)
            case element(tei:table)
                return
                    tei2plain:make-block($node)
            default
                return
                    tei2plain:convert($node/node())
};

(:~
 : This function calls all functions which are needed to convert a TEI-XML to plain text.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $doc: A TEI-Document.
 : @param $fassung: The Fassung that should be converted to plain text.
:)
declare function tei2plain:complete-corr($doc as node()*, $fassung as xs:string) {
    let $single := simplify:all($doc, $fassung)
    let $removed := removal:removal($single)
    let $removedsic := removal:delete-sic($removed)
    let $qm := qm:all($removedsic)
    let $plain := tei2plain:convert($qm)
    return
        $plain
};

(:~
 : This function calls all functions which are needed to convert a TEI-XML to plain text.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $doc: A TEI-Document.
 : @param $fassung: The Fassung that should be converted to plain text.
:)
declare function tei2plain:complete-sic($doc as node()*, $fassung as xs:string) {
    simplify:all($doc, $fassung)
        => removal:removal()
        => removal:delete-corr()
        => qm:all()
        => tei2plain:convert()
};